import{_ as e,l as o,H as i,e as n,I as p}from"./Mermaid-BBeqlX7u.js";import{p as g}from"./treemap-KMMF4GRG-BR4RCA6N.js";import"./index-BPc98LNf.js";import"./_commonjsHelpers-CqkleIqs.js";import"./purify.es-HRjpPm7y.js";import"./min-DbPhMLag.js";import"./_baseUniq-BULAvS64.js";var v={parse:e(async r=>{const a=await g("info",r);o.debug(a)},"parse")},d={version:p.version+""},m=e(()=>d.version,"getVersion"),c={getVersion:m},l=e((r,a,s)=>{o.debug(`rendering info diagram
`+r);const t=i(a);n(t,100,400,!0),t.append("g").append("text").attr("x",100).attr("y",40).attr("class","version").attr("font-size",32).style("text-anchor","middle").text(`v${s}`)},"draw"),f={draw:l},D={parser:v,db:c,renderer:f};export{D as diagram};
