.datagrid.svelte-13v1on6{padding:.75rem 1rem}.datagrid.compact.svelte-13v1on6{padding:.5rem}.datagrid-toolbar.svelte-13v1on6{display:flex;justify-content:space-between;align-items:center;margin-bottom:.5rem;gap:.5rem}.datagrid-search.svelte-13v1on6{flex:1;padding:.4rem .6rem;border:1px solid var(--color-border, #f0ebe4);border-radius:var(--radius-lg, 12px);background:var(--color-card-bg, #fcfaf7);color:var(--color-text, #1f1f1f);font-size:.875rem}.export-btn.svelte-13v1on6{padding:.35rem .75rem;border-radius:var(--radius-lg, 12px);border:1px solid var(--color-border, #f0ebe4);background:var(--color-pill-ghost-bg, #f4f0ea);color:var(--color-text, #1f1f1f);font-size:.75rem;cursor:pointer}.table-wrapper.svelte-13v1on6{overflow-x:auto}table.svelte-13v1on6{width:100%;border-collapse:collapse;font-size:.875rem}th.svelte-13v1on6,td.svelte-13v1on6{padding:.5rem .75rem;border-bottom:1px solid var(--color-border, #f0ebe4)}th.sortable.svelte-13v1on6{cursor:pointer;-webkit-user-select:none;user-select:none;color:var(--color-text, #1f1f1f)}.sort-indicator.svelte-13v1on6{margin-left:.25rem;font-size:.65rem}.striped.svelte-13v1on6{background:var(--color-code-bg, #fbf8f3)}.align-left.svelte-13v1on6{text-align:left}.align-right.svelte-13v1on6{text-align:right}.align-center.svelte-13v1on6{text-align:center}.select-col.svelte-13v1on6{width:36px}.pagination.svelte-13v1on6{display:flex;justify-content:center;align-items:center;gap:.75rem;margin-top:.5rem;font-size:.75rem}.pagination.svelte-13v1on6 button:where(.svelte-13v1on6){border:1px solid var(--color-border, #f0ebe4);background:var(--color-card-bg, #fcfaf7);color:var(--color-text, #1f1f1f);padding:.25rem .5rem;border-radius:.25rem;cursor:pointer}.pagination.svelte-13v1on6 button:where(.svelte-13v1on6):disabled{cursor:not-allowed;opacity:.5}
