import{p as E,a as g,s as k,b as G,H,d as c,f as w,j as R,e as v,h as q,i as b,g as a,t as z,m as B,n as m,G as F,c as J,D as K,A as L,o as N,r as P,x as Q}from"./index-BPc98LNf.js";import T from"./Markdown-BCaT_MOr.js";import"./purify.es-HRjpPm7y.js";import"./katex-XbL3y5x-.js";/* empty css                  */var U=w('<details class="svelte-1b7h7k8"><summary class="svelte-1b7h7k8"> </summary> <div class="panel svelte-1b7h7k8"><!></div></details>'),V=w('<div class="accordion svelte-1b7h7k8"></div>');function ee(y,i){E(i,!0);let h=g(i,"items",19,()=>[]),p=g(i,"allowMultiple",3,!1),r=k(null),n=k(G(new Set));H(()=>{const t=new Set;if(h().forEach((e,o)=>{e.defaultOpen&&t.add(o)}),p())c(n,t,!0);else{const[e]=Array.from(t);c(r,e??null,!0)}});function M(t){if(p()){const e=new Set(a(n));e.has(t)?e.delete(t):e.add(t),c(n,e,!0);return}c(r,a(r)===t?null:t,!0)}function A(t){return p()?a(n).has(t):a(r)===t}var _=V();R(_,21,h,N,(t,e,o)=>{var u=U(),d=v(u);d.__click=s=>{s.preventDefault(),M(o)};var S=v(d),x=q(d,2),D=v(x);{var I=s=>{{let l=J(()=>a(e).props??{});F(s,{get component(){return a(e).component},get props(){return a(l)}})}},O=s=>{var l=K(),j=L(l);{var C=f=>{T(f,{get content(){return a(e).content},padded:!1})};b(j,f=>{a(e).content&&f(C)},!0)}m(s,l)};b(D,s=>{a(e).component?s(I):s(O,!1)})}z(s=>{u.open=s,B(S,a(e).title)},[()=>A(o)]),m(t,u)}),m(y,_),P()}Q(["click"]);export{ee as default};
