import{a as e,f,t as o,k as n,F as a,g as b,c as u,n as v}from"./index-BPc98LNf.js";var g=f('<iframe class="html-frame svelte-1cbd4lv" title="HTML preview"></iframe>');function y(l,t){let d=e(t,"html",3,""),i=e(t,"css",3,void 0),r=e(t,"js",3,void 0),m=e(t,"height",3,"400px"),c=e(t,"sandbox",3,"allow-scripts");const h=u(()=>`<!DOCTYPE html><html><head><style>${i()??""}</style></head><body>${d()??""}<script>${r()??""}<\/script></body></html>`);var s=g();o(()=>{n(s,`height: ${m()};`),a(s,"sandbox",c()),a(s,"srcdoc",b(h))}),v(l,s)}export{y as default};
