import{p as q,a as r,f as i,i as h,e,h as b,g as u,s as z,t as _,l as A,n as v,r as B,d as D,m as f,x as E}from"./index-BPc98LNf.js";import F from"./Markdown-BCaT_MOr.js";import"./purify.es-HRjpPm7y.js";import"./katex-XbL3y5x-.js";/* empty css                  */var G=i('<button class="callout-header svelte-y8yhe6" type="button"><span class="callout-title"> </span> <span class="callout-toggle"> </span></button>'),I=i('<div class="callout-header svelte-y8yhe6"><span class="callout-title"> </span></div>'),J=i('<div class="callout-body svelte-y8yhe6"><!></div>'),K=i("<div><!> <!></div>");function R(g,s){q(s,!0);let x=r(s,"content",3,""),n=r(s,"type",3,"info"),m=r(s,"title",3,void 0),k=r(s,"collapsible",3,!1),o=z(!0);var c=K(),y=e(c);{var C=a=>{var t=G();t.__click=()=>D(o,!u(o));var l=e(t),p=e(l),d=b(l,2),M=e(d);_(j=>{f(p,j),f(M,u(o)?"−":"+")},[()=>m()||n().toUpperCase()]),v(a,t)},w=a=>{var t=I(),l=e(t),p=e(l);_(d=>f(p,d),[()=>m()||n().toUpperCase()]),v(a,t)};h(y,a=>{k()?a(C):a(w,!1)})}var U=b(y,2);{var H=a=>{var t=J(),l=e(t);F(l,{get content(){return x()},allowHtml:!1,padded:!1}),v(a,t)};h(U,a=>{u(o)&&a(H)})}_(()=>A(c,1,`callout ${n()}`,"svelte-y8yhe6")),v(g,c),B()}E(["click"]);export{R as default};
