.artifact-report.svelte-djhbyc{max-width:900px;margin:0 auto;padding:clamp(1.25rem,4vw,2.5rem);line-height:1.7;color:var(--color-text-secondary, #3c3a36)}.report-header.svelte-djhbyc{margin-bottom:2rem;padding-bottom:1rem;border-bottom:1px solid var(--color-border, #f0ebe4)}.report-header.svelte-djhbyc h1:where(.svelte-djhbyc){margin:0 0 .5rem;font-size:2rem;color:var(--color-text, #1f1f1f)}.subtitle.svelte-djhbyc{margin:0 0 .5rem;font-size:1.25rem;color:var(--color-muted, #6b665f)}.meta.svelte-djhbyc{display:flex;gap:1rem;font-size:.875rem;color:var(--color-muted-light, #8a847c)}.toc.svelte-djhbyc{margin-bottom:2rem;padding:1rem 1.25rem;background:var(--color-pill-ghost-bg, #f4f0ea);border:1px solid var(--color-border, #f0ebe4);border-radius:var(--radius-xl, 16px)}.toc.svelte-djhbyc h2:where(.svelte-djhbyc){margin:0 0 .5rem;font-size:1rem;color:var(--color-text, #1f1f1f)}.toc.svelte-djhbyc ul:where(.svelte-djhbyc){margin:0;padding-left:1rem;list-style:none}.toc.svelte-djhbyc li:where(.svelte-djhbyc){margin:.25rem 0}.toc.svelte-djhbyc a:where(.svelte-djhbyc){color:var(--color-primary-text, #1f6c68);text-decoration:underline;text-underline-offset:2px}.toc.svelte-djhbyc a:where(.svelte-djhbyc):hover{color:var(--color-primary-light, #106c67);text-decoration-thickness:2px}.report-body.svelte-djhbyc section:where(.svelte-djhbyc){margin-bottom:2rem}.report-body.svelte-djhbyc h2:where(.svelte-djhbyc){margin:0 0 1rem;padding-bottom:.5rem;border-bottom:1px solid var(--color-border, #f0ebe4);color:var(--color-text, #1f1f1f)}.report-body.svelte-djhbyc h3:where(.svelte-djhbyc){margin:1.5rem 0 .75rem;color:var(--color-text, #1f1f1f)}.subsection.svelte-djhbyc{margin-top:1.5rem;padding-top:.25rem}.component-figure.svelte-djhbyc{margin:1.5rem 0}figcaption.svelte-djhbyc{margin-top:.5rem;font-size:.875rem;color:var(--color-muted, #6b665f);text-align:center;font-style:italic}.report-footer.svelte-djhbyc{margin-top:2rem;padding-top:1rem;border-top:1px solid var(--color-border, #f0ebe4);font-size:.875rem;color:var(--color-muted, #6b665f)}
